/*++========================================================================
 * Program Name:     Novell NCP Redirector for Linux
 * File Name:        nwcapi.c
 * Version:          v1.00
 * Author:           James Turner/Richard Williams
 *
 * Abstract:         This module contains functions used to interface to
 *                   the library interface of the daemon.
 * Notes:
 * Revision History: 
 *
 *
 * Copyright (C) 2005 Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *======================================================================--*/

/*===[ Include files specific to Linux ]==================================*/

/*===[ Include files specific to this module ]============================*/
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/timer.h>
#include <linux/poll.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>

#include "nwcapi.h"
#include "nwerror.h"
#include "commands.h"

#include "vfs.h"

/*===[ External data ]====================================================*/

/*===[ External prototypes ]==============================================*/

extern int DbgPrint( char *Fmt, ... );
extern void mydump(int size, void *dumpptr);

extern session_t Scope_Get_SessionId( void *Scope );
extern int Queue_Daemon_Command(void *request, u_long reqlen, void *data, int dlen, void **reply, u_long *replen, int interruptible);

extern int do_login(NclString *Server, NclString *Username, NclString *Password, u_long *lgnId, session_t Session);

void	GetUserData(NwcScanConnInfo *connInfo, PXPLAT_CALL_REQUEST cmd, PXPLAT_CALL_REPLY reply);
void	GetConnData(NwcGetConnInfo *connInfo, PXPLAT_CALL_REQUEST cmd, PXPLAT_CALL_REPLY reply);

/*===[ Manifest constants ]===============================================*/

/*===[ Type definitions ]=================================================*/

/*===[ Function prototypes ]==============================================*/

/*===[ Global variables ]=================================================*/

/*===[ Code ]=============================================================*/

/*++======================================================================*/

/*++======================================================================*/
int NwOpenConnByName(PXPLAT pdata, u_long *Handle, session_t Session)
/*
 *
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
PNwdCOpenConnByName	  	openConn, connReply;
NwcOpenConnByName 		ocbn;
u_long			retCode = 0, cmdlen, datalen, replylen;
char					*data;

	copy_from_user(&ocbn, pdata->reqData, sizeof(ocbn));
	datalen = sizeof(*openConn) + strlen_user(ocbn.pName->pString) + strlen_user(ocbn.pServiceType);
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_OPEN_CONN_BY_NAME;

		cmd->dataLen = datalen;
		openConn = (PNwdCOpenConnByName)cmd->data;					  

		openConn->nameLen = strlen_user(ocbn.pName->pString);
		openConn->serviceLen = strlen_user(ocbn.pServiceType);
		openConn->uConnFlags = ocbn.uConnFlags;
		openConn->ConnHandle = ocbn.ConnHandle;
		data = (char *)openConn;
		data += sizeof(*openConn);
		openConn->oName = sizeof(*openConn);

		openConn->oServiceType = openConn->oName + openConn->nameLen;
		copy_from_user(data, ocbn.pName->pString, openConn->nameLen);
		data += openConn->nameLen;
		copy_from_user(data, ocbn.pServiceType, openConn->serviceLen);

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			/*
			 * we got reply data from the daemon
          */
			connReply = (PNwdCOpenConnByName)reply->data;
			retCode = reply->Reply.ErrorCode;
			if (!retCode)
			{
				/*
				 * we got valid data.
             */
				connReply = (PNwdCOpenConnByName)reply->data;
                ocbn.RetConnHandle = connReply->newConnHandle;
                *Handle = connReply->newConnHandle;
				copy_to_user(pdata->reqData, &ocbn, sizeof(ocbn));
				DbgPrint("New Conn Handle = %X\n", connReply->newConnHandle);
			}
			Novfs_Free(reply);
		}

		Novfs_Free(cmd);
	}

	return((int)retCode);

}

/*++======================================================================*/
int NwOpenConnByAddr(PXPLAT pdata, u_long *Handle, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
PNwdCOpenConnByAddr	  	openConn, connReply;
NwcOpenConnByAddr 		ocba;
NwcTranAddr            tranAddr;
u_long			retCode = 0, cmdlen, datalen, replylen;
char                 addr[MAX_ADDRESS_LENGTH];

	copy_from_user(&ocba, pdata->reqData, sizeof(ocba));
	datalen = sizeof(*openConn);
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand 			= NWC_OPEN_CONN_BY_ADDRESS;
		cmd->dataLen 				= datalen;
		openConn 					= (PNwdCOpenConnByAddr)cmd->data;

		copy_from_user(&tranAddr, ocba.pTranAddr, sizeof(tranAddr));

	   DbgPrint("NwOpenConnByAddr: tranAddr\n");
      mydump(sizeof(tranAddr), &tranAddr);

      openConn->TranAddr.uTransportType = tranAddr.uTransportType;
      openConn->TranAddr.uAddressLength = tranAddr.uAddressLength;
      memset(addr, 0xcc, sizeof(addr)-1);

      copy_from_user(addr, tranAddr.puAddress, tranAddr.uAddressLength);

	   DbgPrint("NwOpenConnByAddr: addr\n");
      mydump(sizeof(addr), addr);
      
      openConn->TranAddr.oAddress = *(u_long *)(&addr[2]);

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			/*
			 * we got reply data from the daemon
          */
			connReply = (PNwdCOpenConnByAddr)reply->data;
			retCode = reply->Reply.ErrorCode;
			if (!retCode)
			{
				/*
				 * we got valid data.
             */
				connReply = (PNwdCOpenConnByAddr)reply->data;
				ocba.ConnHandle = connReply->ConnHandle;
            *Handle = connReply->ConnHandle;
				copy_to_user(pdata->reqData, &ocba, sizeof(ocba));
				DbgPrint("New Conn Handle = %X\n", connReply->ConnHandle);
			}
			Novfs_Free(reply);
		}

		Novfs_Free(cmd);
	}

	return(retCode);

}

/*++======================================================================*/
int NwOpenConnByRef(PXPLAT pdata, u_long *Handle, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
PNwdCOpenConnByRef	  	openConn;
NwcOpenConnByReference 	ocbr;
u_long			retCode = -ENOMEM, cmdlen, datalen, replylen;

	
	copy_from_user(&ocbr, pdata->reqData, sizeof(ocbr));
	datalen = sizeof(*openConn);
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand 			= NWC_OPEN_CONN_BY_REFERENCE;
		cmd->dataLen 				= datalen;
		openConn 					= (PNwdCOpenConnByRef)cmd->data;

		memcpy(openConn, &ocbr, sizeof(ocbr));
		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			/*
			 * we got reply data from the daemon
          */
			openConn = (PNwdCOpenConnByRef)reply->data;
			retCode = reply->Reply.ErrorCode;
			if (!retCode)
			{
				/*
				 * we got valid data.
             */
				ocbr.ConnHandle = openConn->ConnHandle;
            *Handle = openConn->ConnHandle;

				copy_to_user(pdata->reqData, &ocbr, sizeof(ocbr));
				DbgPrint("New Conn Handle = %X\n", openConn->ConnHandle);
			}
			Novfs_Free(reply);
		}

		Novfs_Free(cmd);
	}
	return(retCode);

}

/*++======================================================================*/
int NwRawSend(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
NwcRequest				xRequest;
PNwcFrag				frag, cFrag, reqFrag;
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
u_long			retCode = -ENOMEM, cmdlen, datalen, replylen;
u_long			x, totalLen;
PNwdCNCPReq				ncpData;
PNwdCNCPRep				ncpReply;
u_char			*reqData;

	DbgPrint("[XPLAT] Process Raw NCP Send\n");
	copy_from_user(&xRequest, pdata->reqData, sizeof(xRequest));

	/*
	 * Figure out the length of the request
    */
	frag = Novfs_Malloc(xRequest.uNumReplyFrags * sizeof(NwcFrag), GFP_KERNEL);

	DbgPrint("[XPLAT RawNCP] - Reply Frag Count 0x%X\n", xRequest.uNumReplyFrags);

	if (!frag)
		return(retCode);		

	copy_from_user(frag, xRequest.pReplyFrags, xRequest.uNumReplyFrags * sizeof(NwcFrag));
	totalLen = 0;

	cFrag = frag;
	for (x = 0; x < xRequest.uNumReplyFrags; x ++)
	{
		DbgPrint("[XPLAT - RawNCP] - Frag Len = %d\n", cFrag->uLength);
		totalLen += cFrag->uLength;
		cFrag++;
	}


	DbgPrint("[XPLAT - RawNCP] - totalLen = %d\n", totalLen);
	datalen = 0;
	reqFrag = Novfs_Malloc(xRequest.uNumRequestFrags * sizeof(NwcFrag), GFP_KERNEL);
	if (!reqFrag)
	{
		Novfs_Free(frag);
		return(retCode);
	}

	copy_from_user(reqFrag, xRequest.pRequestFrags, xRequest.uNumRequestFrags * sizeof(NwcFrag));
	cFrag = reqFrag;
	for (x = 0; x < xRequest.uNumRequestFrags; x ++)
	{
		datalen += cFrag->uLength;
		cFrag++;
	}

	/*
	 * Allocate the cmd Request
    */
	cmdlen = datalen + sizeof(*cmd) + sizeof(*ncpData);
	DbgPrint("[XPLAT RawNCP} - Frag Count 0x%X\n", xRequest.uNumRequestFrags);
	DbgPrint("[XPLAT RawNCP] - Total Command Data Len = %x\n", cmdlen);	

	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_RAW_NCP_REQUEST;

		/*
		 * build the NCP Request
       */
		cmd->dataLen = cmdlen - sizeof(*cmd);
		ncpData = (PNwdCNCPReq)cmd->data;
		ncpData->replyLen = totalLen;
		ncpData->requestLen = datalen;
		ncpData->ConnHandle = xRequest.ConnHandle;
		ncpData->function = xRequest.uFunction;


		reqData = ncpData->data;
		cFrag = reqFrag;

		for (x = 0; x < xRequest.uNumRequestFrags; x ++)
		{
			copy_from_user(reqData, cFrag->pData, cFrag->uLength);
			reqData += cFrag->uLength;
			cFrag++;
		}

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		DbgPrint("RawNCP - reply = %x\n", reply);	
		DbgPrint("RawNCP - retCode = %x\n", retCode);	

		if (reply)
		{
			/*
			 * we got reply data from the daemon
          */
			ncpReply = (PNwdCNCPRep)reply->data;
			retCode = reply->Reply.ErrorCode;

			DbgPrint("RawNCP - Reply Frag Count 0x%X\n", xRequest.uNumReplyFrags);

			/*
			 * We need to copy the reply frags to the packet.
          */
			reqData = ncpReply->data;
			cFrag = frag;

			totalLen = ncpReply->replyLen;
			for (x = 0; x < xRequest.uNumReplyFrags; x ++)
			{
		
				DbgPrint("RawNCP - Copy Frag %d: 0x%X\n", x, cFrag->uLength);
						
				copy_to_user(cFrag->pData, reqData, min(cFrag->uLength, totalLen));
				totalLen -= cFrag->uLength;
				reqData += cFrag->uLength;
				cFrag++;
			}

			Novfs_Free(reply);
		}
		
		Novfs_Free(cmd);
	}

	Novfs_Free(reqFrag);
	Novfs_Free(frag);

	return(retCode);		
}

/*++======================================================================*/
int NwConnClose(PXPLAT pdata, u_long *Handle, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
NwcCloseConn			cc;
PNwdCCloseConn			nwdClose;
u_long			retCode = 0, cmdlen, datalen, replylen;

	copy_from_user(&cc, pdata->reqData, sizeof(cc));

	datalen = sizeof(*nwdClose);
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_CLOSE_CONN;

		nwdClose = (PNwdCCloseConn)cmd->data;
		cmd->dataLen = sizeof(*nwdClose);
		nwdClose->ConnHandle = cc.ConnHandle;
      *Handle = cc.ConnHandle;

		/*
		 * send the request
       */
		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, 0);
		if (reply)
		{
			retCode = reply->Reply.ErrorCode;
			Novfs_Free(reply);
		}
		Novfs_Free(cmd);

	}

	return(retCode);		

}

/*++======================================================================*/
int NwSysConnClose(PXPLAT pdata, u_long *Handle, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
NwcCloseConn			cc;
PNwdCCloseConn			nwdClose;
u_long			retCode = 0, cmdlen, datalen, replylen;

	copy_from_user(&cc, pdata->reqData, sizeof(cc));

	datalen = sizeof(*nwdClose);
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_SYS_CLOSE_CONN;

		nwdClose = (PNwdCCloseConn)cmd->data;
		cmd->dataLen = sizeof(*nwdClose);
		nwdClose->ConnHandle = cc.ConnHandle;
      *Handle = cc.ConnHandle;

		/*
		 * send the request
       */
		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, 0);
		if (reply)
		{
			retCode = reply->Reply.ErrorCode;
			Novfs_Free(reply);
		}
		Novfs_Free(cmd);

	}

	return(retCode);		

}

/*++======================================================================*/
int	NwLoginIdentity(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
NwcLoginIdentity	lgn, *plgn;
int retCode = -ENOMEM;
NclString  server;
NclString  username;
NclString  password;

	copy_from_user(&lgn, pdata->reqData, sizeof(lgn));

	if ( (server.buffer = Novfs_Malloc(lgn.pDomainName->DataLen, GFP_KERNEL)) )
	{
		server.type = lgn.pDomainName->DataType;
		copy_from_user(&server.len, &lgn.pDomainName->DataLen, sizeof(server.len));
		if ( !copy_from_user((void *)server.buffer, lgn.pDomainName->pBuffer, server.len) )
		{
			if ( (username.buffer = Novfs_Malloc(lgn.pObjectName->DataLen, GFP_KERNEL)) )
			{
				username.type = lgn.pObjectName->DataType;
				copy_from_user(&username.len, &lgn.pObjectName->DataLen, sizeof(username.len));
				if ( !copy_from_user((void *)username.buffer, lgn.pObjectName->pBuffer, username.len) )
				{
					if ( (password.buffer = Novfs_Malloc(lgn.pPassword->DataLen, GFP_KERNEL)) )
					{
						password.type = lgn.pPassword->DataType;
						copy_from_user(&password.len, &lgn.pPassword->DataLen, sizeof(password.len));
						if ( !copy_from_user((void *)password.buffer, lgn.pPassword->pBuffer, password.len) )
						{
							retCode = do_login(&server, &username, &password, (u_long *)&lgn.AuthenticationId, Session);
                            if (retCode)
                            {
                                lgn.AuthenticationId = 0;
                            }

                            plgn = (NwcLoginIdentity *)pdata->reqData;
                            copy_to_user(&plgn->AuthenticationId, &lgn.AuthenticationId, sizeof(plgn->AuthenticationId));

						}
                  memset(password.buffer, 0, password.len);
						Novfs_Free(password.buffer);
					}
				}
            memset(username.buffer, 0, username.len);
				Novfs_Free(username.buffer);
			}
		}
		Novfs_Free(server.buffer);
	}
	return(retCode);
}

/*++======================================================================*/
int	NwAuthConnWithId(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
NwcAuthenticateWithId 	pauth;
PNwdCAuthenticateWithId pDauth;
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
u_long			retCode = -ENOMEM, cmdlen, datalen, replylen;

	datalen = sizeof(*pDauth);
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_AUTHENTICATE_CONN_WITH_ID;


		copy_from_user(&pauth, pdata->reqData, sizeof(pauth));

		pDauth = (PNwdCAuthenticateWithId)cmd->data;
		cmd->dataLen = datalen;
		pDauth->AuthenticationId = pauth.AuthenticationId;
		pDauth->ConnHandle = pauth.ConnHandle;

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			retCode = reply->Reply.ErrorCode;
			Novfs_Free(reply);
		}
		Novfs_Free(cmd);
	}
	return(retCode);
}

/*++======================================================================*/
int	NwLicenseConn(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
NwcLicenseConn			lisc;
PNwdCLicenseConn		pDLisc;
u_long			retCode = -ENOMEM, cmdlen, datalen, replylen;

	datalen = sizeof(*pDLisc);
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_LICENSE_CONN;


		copy_from_user(&lisc, pdata->reqData, sizeof(lisc));

		pDLisc = (PNwdCLicenseConn)cmd->data;
		cmd->dataLen = datalen;
		pDLisc->ConnHandle = lisc.ConnHandle;

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			retCode = reply->Reply.ErrorCode;
			Novfs_Free(reply);
		}
		Novfs_Free(cmd);
	}
	return(retCode);
}


/*++======================================================================*/
int	NwLogoutIdentity(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
NwcLogoutIdentity		logout;
PNwdCLogoutIdentity		pDLogout;
u_long			retCode = -ENOMEM, cmdlen, datalen, replylen;

	datalen = sizeof(*pDLogout);
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_LOGOUT_IDENTITY;

		copy_from_user(&logout, pdata->reqData, sizeof(logout));

		pDLogout = (PNwdCLogoutIdentity)cmd->data;
		cmd->dataLen = datalen;
		pDLogout->AuthenticationId = logout.AuthenticationId;

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			retCode = reply->Reply.ErrorCode;
			Novfs_Free(reply);
		}
		Novfs_Free(cmd);
	}
	return(retCode);
}

/*++======================================================================*/
int	NwUnlicenseConn(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
PNwdCUnlicenseConn		pUconn;
NwcUnlicenseConn 		ulc;
u_long			retCode = -ENOMEM, cmdlen, datalen, replylen;

	
	copy_from_user(&ulc, pdata->reqData, sizeof(ulc));
	datalen = sizeof(*pUconn);
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand 			= NWC_UNLICENSE_CONN;
		cmd->dataLen 				= datalen;
		pUconn 						= (PNwdCUnlicenseConn)cmd->data;

		pUconn->ConnHandle = ulc.ConnHandle;
		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			/*
			 * we got reply data from the daemon
          */
			retCode = reply->Reply.ErrorCode;
			Novfs_Free(reply);
		}

		Novfs_Free(cmd);
	}
	return(retCode);

}


/*++======================================================================*/
int	NwUnAuthenticate(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
NwcUnauthenticate		auth;
PNwdCUnauthenticate		pDAuth;
u_long			retCode = -ENOMEM, cmdlen, datalen, replylen;

	datalen = sizeof(*pDAuth);
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_UNAUTHENTICATE_CONN;

		copy_from_user(&auth, pdata->reqData, sizeof(auth));

		pDAuth = (PNwdCUnauthenticate)cmd->data;
		cmd->dataLen = datalen;
		pDAuth->AuthenticationId = auth.AuthenticationId;
		pDAuth->ConnHandle = auth.ConnHandle;

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			retCode = reply->Reply.ErrorCode;
			Novfs_Free(reply);
		}
		Novfs_Free(cmd);
	}
	return(retCode);

}


/*++======================================================================*/
int	NwGetConnInfo(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
NwcGetConnInfo			connInfo;
PNwdCGetConnInfo		pDConnInfo;
u_long			retCode = -ENOMEM, cmdlen, replylen;

	cmdlen = sizeof(*cmd) + sizeof(*pDConnInfo);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
	copy_from_user(&connInfo, pdata->reqData, sizeof(NwcGetConnInfo));

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_GET_CONN_INFO;
		
		pDConnInfo = (PNwdCGetConnInfo)cmd->data;

		pDConnInfo->ConnHandle = connInfo.ConnHandle;
		pDConnInfo->uInfoLevel = connInfo.uInfoLevel;
		pDConnInfo->uInfoLength = connInfo.uInfoLength;
		cmd->dataLen = sizeof(*pDConnInfo);

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			retCode = reply->Reply.ErrorCode;
			GetConnData(&connInfo, cmd, reply);

			Novfs_Free(reply);
		}
		Novfs_Free(cmd);

	}

	return(retCode);

}


/*++======================================================================*/
int	NwSetConnInfo(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
NwcSetConnInfo			connInfo;
PNwdCSetConnInfo		pDConnInfo;
u_long			retCode = -ENOMEM, cmdlen, replylen;

	cmdlen = sizeof(*cmd) + sizeof(*pDConnInfo);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
	copy_from_user(&connInfo, pdata->reqData, sizeof(NwcSetConnInfo));

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_SET_CONN_INFO;
		
		pDConnInfo = (PNwdCSetConnInfo)cmd->data;

		pDConnInfo->ConnHandle = connInfo.ConnHandle;
		pDConnInfo->uInfoLevel = connInfo.uInfoLevel;
		pDConnInfo->uInfoLength = connInfo.uInfoLength;
		cmd->dataLen = sizeof(*pDConnInfo);

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			retCode = reply->Reply.ErrorCode;
			Novfs_Free(reply);
		}
		Novfs_Free(cmd);

	}

	return(retCode);

}

/*++======================================================================*/
int	NwGetIdentityInfo(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
NwcGetIdentityInfo		qidInfo, *gId;
PNwdCGetIdentityInfo	idInfo;
NwcString				xferStr;
char					*str;
u_long			retCode = -ENOMEM, cmdlen, replylen;

	cmdlen = sizeof(*cmd) + sizeof(*idInfo);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
	copy_from_user(&qidInfo, pdata->reqData, sizeof(qidInfo));

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_GET_IDENTITY_INFO;
		
		idInfo = (PNwdCGetIdentityInfo)cmd->data;

		idInfo->AuthenticationId = qidInfo.AuthenticationId;
		cmd->dataLen = sizeof(*idInfo);

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			retCode = reply->Reply.ErrorCode;
	
            if (!reply->Reply.ErrorCode)
            {
			    /*
			     * Save the return info to the user structure.
              */
			    gId = pdata->reqData;
			    idInfo = (PNwdCGetIdentityInfo)reply->data;
			    copy_to_user(&gId->AuthenticationId, &idInfo->AuthenticationId, sizeof(idInfo->AuthenticationId));
			    copy_to_user(&gId->AuthType, &idInfo->AuthType, sizeof(idInfo->AuthType));
			    copy_to_user(&gId->IdentityFlags, &idInfo->IdentityFlags, sizeof(idInfo->IdentityFlags));
			    copy_to_user(&gId->NameType, &idInfo->NameType, sizeof(idInfo->NameType));
			    copy_to_user(&gId->ObjectType, &idInfo->ObjectType, sizeof(idInfo->ObjectType));

			    copy_from_user(&xferStr, gId->pDomainName, sizeof(NwcString));
			    str = (char *)((char *)reply->data + idInfo->pDomainNameOffset);
			    copy_to_user(xferStr.pBuffer, str, idInfo->domainLen);
			    xferStr.DataType = NWC_STRING_TYPE_ASCII;
			    xferStr.DataLen = idInfo->domainLen;
			    copy_to_user(gId->pDomainName, &xferStr, sizeof(NwcString));


			    copy_from_user(&xferStr, gId->pObjectName, sizeof(NwcString));
			    str = (char *)((char *)reply->data + idInfo->pObjectNameOffset);
			    copy_to_user(xferStr.pBuffer, str, idInfo->objectLen);
			    xferStr.DataLen = idInfo->objectLen - 1;
			    xferStr.DataType = NWC_STRING_TYPE_ASCII;
			    copy_to_user(gId->pObjectName, &xferStr, sizeof(NwcString));
            }

			Novfs_Free(reply);
		}
		Novfs_Free(cmd);

	}

	return(retCode);
}

/*++======================================================================*/
int	NwScanConnInfo(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST		cmd;
PXPLAT_CALL_REPLY		reply;
NwcScanConnInfo			connInfo, *rInfo;
PNwdCScanConnInfo		pDConnInfo;
u_long			retCode = -ENOMEM, cmdlen, replylen;
u_char 			*localData;

	copy_from_user(&connInfo, pdata->reqData, sizeof(NwcScanConnInfo));

	cmdlen = sizeof(*cmd) + sizeof(*pDConnInfo) + connInfo.uScanInfoLen;
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_SCAN_CONN_INFO;
		
		pDConnInfo = (PNwdCScanConnInfo)cmd->data;

		DbgPrint("NwScanConnInfo: Input Data\n");
		DbgPrint("connInfo.uScanIndex = 0x%X\n", connInfo.uScanIndex);
		DbgPrint("connInfo.uConnectionReference = 0x%X\n", connInfo.uConnectionReference);
		DbgPrint("connInfo.uScanInfoLevel = 0x%X\n", connInfo.uScanInfoLevel);
		DbgPrint("connInfo.uScanInfoLen = 0x%X\n", connInfo.uScanInfoLen);
		DbgPrint("connInfo.uReturnInfoLength = 0x%X\n", connInfo.uReturnInfoLength);
		DbgPrint("connInfo.uReturnInfoLevel = 0x%X\n", connInfo.uReturnInfoLevel);
		DbgPrint("connInfo.uScanFlags = 0x%X\n", connInfo.uScanFlags);


		pDConnInfo->uScanIndex = connInfo.uScanIndex;
		pDConnInfo->uConnectionReference = connInfo.uConnectionReference;
		pDConnInfo->uScanInfoLevel = connInfo.uScanInfoLevel;
		pDConnInfo->uScanInfoLen = connInfo.uScanInfoLen;
		pDConnInfo->uReturnInfoLength = connInfo.uReturnInfoLength;
		pDConnInfo->uReturnInfoLevel = connInfo.uReturnInfoLevel;
		pDConnInfo->uScanFlags = connInfo.uScanFlags;

		if (pDConnInfo->uScanInfoLen)
		{
			localData = (u_char *)pDConnInfo;
			pDConnInfo->uScanConnInfoOffset = sizeof(*pDConnInfo);
			localData += pDConnInfo->uScanConnInfoOffset;
			copy_from_user(localData, connInfo.pScanConnInfo, connInfo.uScanInfoLen);
		}
		else
		{
			pDConnInfo->uScanConnInfoOffset = 0;
		}
			

		cmd->dataLen = sizeof(*pDConnInfo);

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			DbgPrint("NwScanConnInfo: Reply recieved\n");
			DbgPrint("   NextIndex = %x\n", connInfo.uScanIndex);
			DbgPrint("   ErrorCode = %x\n", reply->Reply.ErrorCode);
			DbgPrint("   data = %x\n", reply->data);

			pDConnInfo = (PNwdCScanConnInfo)reply->data;
			retCode = (u_long)reply->Reply.ErrorCode;
			if (!retCode)
			{
				GetUserData(&connInfo, cmd, reply);
				rInfo = (NwcScanConnInfo *)pdata->reqData;
				copy_to_user(pdata->reqData, &pDConnInfo->uScanIndex, sizeof(pDConnInfo->uScanIndex));
				copy_to_user(&rInfo->uConnectionReference, &pDConnInfo->uConnectionReference, sizeof(pDConnInfo->uConnectionReference));
			}
			else
			{
			u_long x;

				x = 0;
				rInfo = (NwcScanConnInfo *)pdata->reqData;
				copy_to_user(&rInfo->uConnectionReference, &x, sizeof(rInfo->uConnectionReference));
			}
				
			Novfs_Free(reply);
		}
		Novfs_Free(cmd);

	}

	return(retCode);
}

/*++======================================================================*/
void	GetUserData(NwcScanConnInfo *connInfo, PXPLAT_CALL_REQUEST cmd, PXPLAT_CALL_REPLY reply)
/*
 *  Abstract:  Copies the user data out of the scan conn info call.
 *
 *========================================================================*/

{
u_long			uLevel;
PNwdCScanConnInfo		pDConnInfo;

u_char			*srcData;
u_long			dataLen = 0;

	
	pDConnInfo = (PNwdCScanConnInfo)reply->data;
	uLevel = pDConnInfo->uScanInfoLevel;
	DbgPrint("[GetUserData] uLevel = %d, reply = 0x%X, reply->data = 0x%X\n", uLevel, reply, reply->data);

	switch(uLevel)
	{
		case NWC_CONN_INFO_RETURN_ALL:
			srcData = (u_char *)reply->data;
			dataLen = reply->dataLen;
		break;

		case NWC_CONN_INFO_RETURN_NONE:
			dataLen = 0;
		break;

		case NWC_CONN_INFO_TRAN_ADDR:
		case NWC_CONN_INFO_NDS_STATE:
		case NWC_CONN_INFO_MAX_PACKET_SIZE:
		case NWC_CONN_INFO_LICENSE_STATE:
		case NWC_CONN_INFO_PUBLIC_STATE:
		case NWC_CONN_INFO_SERVICE_TYPE:
		case NWC_CONN_INFO_DISTANCE:
		case NWC_CONN_INFO_SERVER_VERSION:
		case NWC_CONN_INFO_AUTH_ID:
		case NWC_CONN_INFO_SUSPENDED:
		case NWC_CONN_INFO_WORKGROUP_ID:
		case NWC_CONN_INFO_SECURITY_STATE:
		case NWC_CONN_INFO_CONN_NUMBER:
		case NWC_CONN_INFO_USER_ID:
		case NWC_CONN_INFO_BCAST_STATE:
		case NWC_CONN_INFO_CONN_REF:
		case NWC_CONN_INFO_AUTH_STATE:
		case NWC_CONN_INFO_TREE_NAME:
		case NWC_CONN_INFO_SERVER_NAME:
		case NWC_CONN_INFO_VERSION:
			srcData = (u_char *)pDConnInfo;
			srcData += pDConnInfo->uReturnConnInfoOffset;
			dataLen = pDConnInfo->uReturnInfoLength;
		break;

		case NWC_CONN_INFO_TREE_NAME_UNICODE:
		case NWC_CONN_INFO_SERVER_NAME_UNICODE:
		break;

		case NWC_CONN_INFO_LOCAL_TRAN_ADDR:
		break;

		case NWC_CONN_INFO_ALTERNATE_ADDR:
		break;

		case NWC_CONN_INFO_SERVER_GUID:
		break;

		default:
		break;
	}

	if (dataLen)
	{
		DbgPrint("Copy Data in GetUserData 0x%X -> 0x%X :: 0x%X\n",
						srcData, connInfo->pReturnConnInfo, dataLen);			
		copy_to_user(connInfo->pReturnConnInfo, srcData, dataLen);
	}

	return;
}

/*++======================================================================*/
void	GetConnData(NwcGetConnInfo *connInfo, PXPLAT_CALL_REQUEST cmd, PXPLAT_CALL_REPLY reply)
/*
 *  Abstract:  Copies the user data out of the scan conn info call.
 *
 *========================================================================*/

{
u_long			uLevel;
PNwdCGetConnInfo		pDConnInfo;

u_char			*srcData;
u_long			dataLen = 0;

	
	pDConnInfo = (PNwdCGetConnInfo)cmd->data;
	uLevel = pDConnInfo->uInfoLevel;

	switch(uLevel)
	{
		case NWC_CONN_INFO_RETURN_ALL:
			srcData = (u_char *)reply->data;
			dataLen = reply->dataLen;
		break;

		case NWC_CONN_INFO_RETURN_NONE:
			dataLen = 0;
		break;

		case NWC_CONN_INFO_TRAN_ADDR:
      {
         u_char *dstData = connInfo->pConnInfo;
         NwcTranAddr tranAddr;

         srcData = (u_char *)reply->data;

         copy_from_user(&tranAddr, dstData, sizeof(tranAddr));
         tranAddr.uTransportType = ((PNwdTranAddr)srcData)->uTransportType;
         tranAddr.uAddressLength = ((PNwdTranAddr)srcData)->uAddressLength;
         copy_to_user(dstData, &tranAddr, sizeof(tranAddr));
         copy_to_user(tranAddr.puAddress, ((PNwdTranAddr)srcData)->Buffer, ((PNwdTranAddr)srcData)->uAddressLength);
         dataLen=0;
         break;
      }
		case NWC_CONN_INFO_NDS_STATE:
		case NWC_CONN_INFO_MAX_PACKET_SIZE:
		case NWC_CONN_INFO_LICENSE_STATE:
		case NWC_CONN_INFO_PUBLIC_STATE:
		case NWC_CONN_INFO_SERVICE_TYPE:
		case NWC_CONN_INFO_DISTANCE:
		case NWC_CONN_INFO_SERVER_VERSION:
		case NWC_CONN_INFO_AUTH_ID:
		case NWC_CONN_INFO_SUSPENDED:
		case NWC_CONN_INFO_WORKGROUP_ID:
		case NWC_CONN_INFO_SECURITY_STATE:
		case NWC_CONN_INFO_CONN_NUMBER:
		case NWC_CONN_INFO_USER_ID:
		case NWC_CONN_INFO_BCAST_STATE:
		case NWC_CONN_INFO_CONN_REF:
		case NWC_CONN_INFO_AUTH_STATE:
		case NWC_CONN_INFO_VERSION:
		case NWC_CONN_INFO_SERVER_NAME:
		case NWC_CONN_INFO_TREE_NAME:
			srcData = (u_char *)reply->data;
			dataLen = reply->dataLen;
		break;

		case NWC_CONN_INFO_TREE_NAME_UNICODE:
		case NWC_CONN_INFO_SERVER_NAME_UNICODE:
		break;

		case NWC_CONN_INFO_LOCAL_TRAN_ADDR:
		break;

		case NWC_CONN_INFO_ALTERNATE_ADDR:
		break;

		case NWC_CONN_INFO_SERVER_GUID:
		break;

		default:
		break;
	}

	if (dataLen)
	{
		copy_to_user(connInfo->pConnInfo, srcData, connInfo->uInfoLength);
	}

	return;
}

/*++======================================================================*/
int	 NwGetDaemonVersion(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/

{
PXPLAT_CALL_REQUEST			cmd;
PXPLAT_CALL_REPLY			reply;
PNwdCGetRequesterVersion	pDVersion;
u_long				retCode = -ENOMEM, cmdlen, datalen, replylen;

	datalen = sizeof(*pDVersion);
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_GET_REQUESTER_VERSION;
		cmdlen = sizeof(*cmd);
		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			retCode = reply->Reply.ErrorCode;
			pDVersion = (PNwdCGetRequesterVersion)reply->data;
			copy_to_user(pDVersion, pdata->reqData, sizeof(*pDVersion));
			Novfs_Free(reply);
		}
		Novfs_Free(cmd);
	}
	return(retCode);

}


/*++======================================================================*/
int	 NwcGetPreferredDSTree(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/

{
PXPLAT_CALL_REQUEST			cmd;
PXPLAT_CALL_REPLY			reply;
PNwdCGetPreferredDsTree		pDGetTree;
NwcGetPreferredDsTree		xplatCall, *p;
u_long				retCode = -ENOMEM, cmdlen, datalen, replylen;
u_char				*dPtr;

	copy_from_user(&xplatCall, pdata->reqData, sizeof(NwcGetPreferredDsTree));
	datalen = sizeof(*pDGetTree) + xplatCall.uTreeLength;
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_GET_PREFERRED_DS_TREE;
		cmdlen = sizeof(*cmd);

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			retCode = reply->Reply.ErrorCode;
			if (!retCode)
			{
				pDGetTree = (PNwdCGetPreferredDsTree)reply->data;
				dPtr = reply->data + pDGetTree->DsTreeNameOffset;
				p = (NwcGetPreferredDsTree *)pdata->reqData;

				DbgPrint("NwcGetPreferredDSTree: Reply recieved\n");
				DbgPrint("   TreeLen = %x\n", pDGetTree->uTreeLength);
				DbgPrint("   TreeName = %s\n", dPtr);

				copy_to_user(p, &pDGetTree->uTreeLength, 4);
				copy_to_user(xplatCall.pDsTreeName, dPtr, pDGetTree->uTreeLength);
			}
         Novfs_Free(reply);
		}
		Novfs_Free(cmd);
	}
	return(retCode);

}

/*++======================================================================*/
int	 NwcSetPreferredDSTree(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/

{
PXPLAT_CALL_REQUEST			cmd;
PXPLAT_CALL_REPLY			reply;
PNwdCSetPreferredDsTree		pDSetTree;
NwcSetPreferredDsTree		xplatCall;
u_long				retCode = -ENOMEM, cmdlen, datalen, replylen;
u_char				*dPtr;

	copy_from_user(&xplatCall, pdata->reqData, sizeof(NwcSetPreferredDsTree));
	datalen = sizeof(*pDSetTree) + xplatCall.uTreeLength;
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_SET_PREFERRED_DS_TREE;

		pDSetTree = (PNwdCSetPreferredDsTree)cmd->data;
		pDSetTree->DsTreeNameOffset = sizeof(*pDSetTree);
		pDSetTree->uTreeLength = xplatCall.uTreeLength;

		dPtr = cmd->data + sizeof(*pDSetTree);
		copy_from_user(dPtr, xplatCall.pDsTreeName, xplatCall.uTreeLength);
		

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
         retCode = reply->Reply.ErrorCode;
         Novfs_Free(reply);
		}
		Novfs_Free(cmd);
	}
	return(retCode);

}


/*++======================================================================*/
int	 NwcSetDefaultNameCtx(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/

{
PXPLAT_CALL_REQUEST			cmd;
PXPLAT_CALL_REPLY			reply;
NwcSetDefaultNameContext	xplatCall;
PNwdCSetDefaultNameContext	pDSet;
u_long				retCode = -ENOMEM, cmdlen, datalen, replylen;
u_char				*dPtr;

	copy_from_user(&xplatCall, pdata->reqData, sizeof(NwcSetDefaultNameContext));
	datalen = sizeof(*pDSet) + xplatCall.uNameLength;
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_SET_DEFAULT_NAME_CONTEXT;

		pDSet = (PNwdCSetDefaultNameContext)cmd->data;
		pDSet->NameContextOffset = sizeof(*pDSet);
		pDSet->uNameLength = xplatCall.uNameLength;

		dPtr = cmd->data + sizeof(*pDSet);
		copy_from_user(dPtr, xplatCall.pNameContext, xplatCall.uNameLength);

        pDSet->TreeOffset = pDSet->NameContextOffset + pDSet->uNameLength + 1;
		dPtr = cmd->data + pDSet->TreeOffset;
		copy_from_user(dPtr, xplatCall.pTreeName, xplatCall.uTreeLength);
		

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
            retCode = reply->Reply.ErrorCode;
            Novfs_Free(reply);
		}
		Novfs_Free(cmd);
	}
	return(retCode);

}

/*++======================================================================*/
int	 NwcGetDefaultNameCtx(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/

{
PXPLAT_CALL_REQUEST			cmd;
PXPLAT_CALL_REPLY			reply;
NwcGetDefaultNameContext	xplatCall;
u_long				retCode = -ENOMEM, cmdlen, replylen;

	copy_from_user(&xplatCall, pdata->reqData, sizeof(NwcGetDefaultNameContext));
	cmdlen = xplatCall.uTreeLength + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_GET_DEFAULT_NAME_CONTEXT;
      cmd->dataLen = xplatCall.uTreeLength;
      copy_from_user( cmd->data, xplatCall.pTreeName, xplatCall.uTreeLength);
      cmd->data[cmd->dataLen] = 0;

		retCode = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
		
			retCode = reply->Reply.ErrorCode;
			if (!retCode)
			{
                pdata->repLen = reply->dataLen;
                DbgPrint("NwcGetDefaultNameCtx: retCode=0x%x uNameLength1=%d uNameLength2=%d\n", retCode, reply->dataLen, xplatCall.uNameLength);
                if (xplatCall.uNameLength < reply->dataLen)
                {
                    reply->dataLen = xplatCall.uNameLength;
                    retCode = NWE_BUFFER_OVERFLOW;
                }
                copy_to_user(xplatCall.pNameContext, reply->data, reply->dataLen);
            }

		    Novfs_Free(reply);
		}
		Novfs_Free(cmd);
	}
	return(retCode);

}

/*++======================================================================*/
int NwQueryFeature(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
   NwcQueryFeature xpCall;
   int status = SUCCESS;

   copy_from_user(&xpCall, pdata->reqData, sizeof(NwcQueryFeature));
   switch (xpCall.Feature)   
   {
      case NWC_FEAT_NDS:
      case NWC_FEAT_NDS_MTREE:
      case NWC_FEAT_PRN_CAPTURE:
      case NWC_FEAT_NDS_RESOLVE:

      status = NWE_REQUESTER_FAILURE;  

   }
   return( status );
}

/*++======================================================================*/
int	NwcGetTreeMonitoredConn(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST				cmd;
PXPLAT_CALL_REPLY				reply;
NwcGetTreeMonitoredConnRef		xplatCall, *p;
PNwdCGetTreeMonitoredConnRef	pDConnRef;
char							*dPtr;
u_long					status = -ENOMEM, cmdlen, datalen, replylen;

	copy_from_user(&xplatCall, pdata->reqData, sizeof(NwcGetTreeMonitoredConnRef));
	datalen = sizeof(*pDConnRef) + xplatCall.pTreeName->DataLen;
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_GET_TREE_MONITORED_CONN_REF;

		pDConnRef = (PNwdCGetTreeMonitoredConnRef)cmd->data;
		pDConnRef->TreeName.boffset = sizeof(*pDConnRef);
		pDConnRef->TreeName.len = xplatCall.pTreeName->DataLen;
		pDConnRef->TreeName.type = xplatCall.pTreeName->DataType;

		dPtr = cmd->data + sizeof(*pDConnRef);
		copy_from_user(dPtr, xplatCall.pTreeName->pBuffer, pDConnRef->TreeName.len);
		status = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			pDConnRef = (PNwdCGetTreeMonitoredConnRef)reply->data;
			dPtr = reply->data + pDConnRef->TreeName.boffset;
			p = (NwcGetTreeMonitoredConnRef *)pdata->reqData;
			copy_to_user(&p->uConnReference, &pDConnRef->uConnReference, 4);

			status = reply->Reply.ErrorCode;
         Novfs_Free(reply);
		}
		Novfs_Free(cmd);

	}

	return(status);
}

/*++======================================================================*/
int	NwcEnumIdentities(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract:
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST				cmd;
PXPLAT_CALL_REPLY				reply;
NwcEnumerateIdentities			xplatCall, *eId;
PNwdCEnumerateIdentities		pEnum;
NwcString						xferStr;
char							*str;
u_long					status = -ENOMEM, cmdlen, datalen, replylen;

	copy_from_user(&xplatCall, pdata->reqData, sizeof(NwcEnumerateIdentities));
	datalen = sizeof(*pEnum);
	cmdlen = datalen + sizeof(*cmd);
	cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

	if (cmd)
	{
		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_ENUMERATE_IDENTITIES;

		DbgPrint("NwcEnumIdentities: Send Request\n");
		DbgPrint("   iterator = %x\n", xplatCall.Iterator);
		DbgPrint("   cmdlen = %d\n", cmdlen);

		pEnum = (PNwdCEnumerateIdentities)cmd->data;
		pEnum->Iterator = xplatCall.Iterator;
		status = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
		if (reply)
		{
			status = reply->Reply.ErrorCode;
			
			eId = pdata->repData;
			pEnum = (PNwdCEnumerateIdentities)reply->data;
			copy_to_user(&eId->Iterator, &pEnum->Iterator, sizeof(pEnum->Iterator));
			DbgPrint("[XPLAT NWCAPI] Found AuthId 0x%X\n", pEnum->AuthenticationId);
			copy_to_user(&eId->AuthenticationId, &pEnum->AuthenticationId, sizeof(pEnum->AuthenticationId));
			copy_to_user(&eId->AuthType, &pEnum->AuthType, sizeof(pEnum->AuthType));
			copy_to_user(&eId->IdentityFlags, &pEnum->IdentityFlags, sizeof(pEnum->IdentityFlags));
			copy_to_user(&eId->NameType, &pEnum->NameType, sizeof(pEnum->NameType));
			copy_to_user(&eId->ObjectType, &pEnum->ObjectType, sizeof(pEnum->ObjectType));

			if (!status)
			{
				copy_from_user(&xferStr, eId->pDomainName, sizeof(NwcString));
				str = (char *)((char *)reply->data + pEnum->domainNameOffset);
				DbgPrint("[XPLAT NWCAPI] Found Domain %s\n", str);
				copy_to_user(xferStr.pBuffer, str, pEnum->domainNameLen);
				xferStr.DataType = NWC_STRING_TYPE_ASCII;
				xferStr.DataLen = pEnum->domainNameLen - 1;
				copy_to_user(eId->pDomainName, &xferStr, sizeof(NwcString));


				copy_from_user(&xferStr, eId->pObjectName, sizeof(NwcString));
				str = (char *)((char *)reply->data + pEnum->objectNameOffset);
				DbgPrint("[XPLAT NWCAPI] Found User %s\n", str);
				copy_to_user(xferStr.pBuffer, str, pEnum->objectNameLen);
				xferStr.DataType = NWC_STRING_TYPE_ASCII;
				xferStr.DataLen = pEnum->objectNameLen - 1;
				copy_to_user(eId->pObjectName, &xferStr, sizeof(NwcString));
			}

		    Novfs_Free(reply);

		}
		Novfs_Free(cmd);

	}
	return(status);
}

/*++======================================================================*/
int	NwcChangeAuthKey(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract: Change the password on the server
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST				cmd;
PXPLAT_CALL_REPLY				reply;
NwcChangeKey			        xplatCall;
PNwdCChangeKey		            pNewKey;
NwcString						xferStr;
char							*str;
u_long                          status = -ENOMEM, cmdlen, datalen, replylen;

	copy_from_user(&xplatCall, pdata->reqData, sizeof(NwcChangeKey));

    datalen = sizeof(NwdCChangeKey) + xplatCall.pDomainName->DataLen + xplatCall.pObjectName->DataLen
        + xplatCall.pNewPassword->DataLen + xplatCall.pVerifyPassword->DataLen;

    cmdlen = sizeof(*cmd) + datalen;
    cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

    if (cmd)
    {
        pNewKey = (PNwdCChangeKey)cmd->data;
        cmd->dataLen = datalen;
   		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_CHANGE_KEY;

        pNewKey->NameType = xplatCall.NameType;
        pNewKey->ObjectType = xplatCall.ObjectType;
        pNewKey->AuthType = xplatCall.AuthType;
        str = (char *)pNewKey;

        /*
         * Get the tree name
         */
        str += sizeof(*pNewKey);
        copy_from_user(&xferStr, xplatCall.pDomainName, sizeof(NwcString));
        pNewKey->domainNameOffset = sizeof(*pNewKey);
        copy_from_user(str, xferStr.pBuffer, xferStr.DataLen);
        pNewKey->domainNameLen = xferStr.DataLen;

        /*
         * Get the User Name
         */
        str += pNewKey->domainNameLen;
        copy_from_user(&xferStr, xplatCall.pObjectName, sizeof(NwcString));
        pNewKey->objectNameOffset = pNewKey->domainNameOffset + pNewKey->domainNameLen;
        copy_from_user(str, xferStr.pBuffer, xferStr.DataLen);
        pNewKey->objectNameLen = xferStr.DataLen;

        /*
         * Get the New Password
         */
        str += pNewKey->objectNameLen;
        copy_from_user(&xferStr, xplatCall.pNewPassword, sizeof(NwcString));
        pNewKey->newPasswordOffset = pNewKey->objectNameOffset + pNewKey->objectNameLen;
        copy_from_user(str, xferStr.pBuffer, xferStr.DataLen);
        pNewKey->newPasswordLen = xferStr.DataLen;

        /*
         * Get the Verify Password
         */
        str += pNewKey->newPasswordLen;
        copy_from_user(&xferStr, xplatCall.pVerifyPassword, sizeof(NwcString));
        pNewKey->verifyPasswordOffset = pNewKey->newPasswordOffset + pNewKey->newPasswordLen;
        copy_from_user(str, xferStr.pBuffer, xferStr.DataLen);
        pNewKey->verifyPasswordLen = xferStr.DataLen;

		status = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
        if (reply)
        {
            status = reply->Reply.ErrorCode;
            Novfs_Free(reply);
        }
        memset(cmd, 0, cmdlen);

        Novfs_Free(cmd);
    }

	return(status);
}

/*++======================================================================*/
int	NwcSetPrimaryConn(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract: Set the primary connection Id
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST        cmd;
PXPLAT_CALL_REPLY          reply;
NwcSetPrimaryConnection    xplatCall;
PNwdCSetPrimaryConnection  pConn;
u_long                     status = -ENOMEM, cmdlen, datalen, replylen;

   copy_from_user(&xplatCall, pdata->reqData, sizeof(NwcSetPrimaryConnection));

   datalen = sizeof(NwdCSetPrimaryConnection);
   cmdlen = sizeof(*cmd) + datalen;
   cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
   if (cmd)
   {
      pConn = (PNwdCSetPrimaryConnection)cmd->data;
      cmd->dataLen = datalen;
      cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
      cmd->Command.SequenceNumber = 0;
      cmd->Command.SessionId      = Session;
      cmd->NwcCommand = NWC_SET_PRIMARY_CONN;
      pConn->ConnHandle = xplatCall.ConnHandle;
      status = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);

      if (reply)
      {
         status = reply->Reply.ErrorCode;
         Novfs_Free(reply);
      }

      Novfs_Free(cmd);
   }

   return(status);
}

/*++======================================================================*/
int	NwcGetPrimaryConn(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract: Get the Primary connection 
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
XPLAT_CALL_REQUEST   cmd;
PXPLAT_CALL_REPLY    reply;
u_long               status = -ENOMEM, cmdlen, replylen;


   cmdlen = (u_long)(&((PXPLAT_CALL_REQUEST)0)->data);

   cmd.dataLen = 0;
   cmd.Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
   cmd.Command.SequenceNumber = 0;
   cmd.Command.SessionId      = Session;
   cmd.NwcCommand = NWC_GET_PRIMARY_CONN;

   status = Queue_Daemon_Command((void *)&cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);

   if (reply)
   {
      status = reply->Reply.ErrorCode;
      if (!status)
      {
         copy_to_user(pdata->repData, reply->data, sizeof(u_long));
      }

      Novfs_Free(reply);
   }

   return(status);
}


/*++======================================================================*/
int	NwcSetMapDrive(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract: Get the Primary connection 
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{

PXPLAT_CALL_REQUEST     cmd;
PXPLAT_CALL_REPLY       reply;
u_long                  status = 0, datalen, cmdlen, replylen;
NwcMapDriveEx           symInfo;

    DbgPrint("Call to NwcSetMapDrive\n");
    copy_from_user(&symInfo, pdata->reqData, sizeof(symInfo));
    cmdlen = sizeof(*cmd);
    datalen = sizeof(symInfo) + symInfo.dirPathOffsetLength + symInfo.linkOffsetLength;

    DbgPrint(" cmdlen = %d\n", cmdlen);
    DbgPrint(" dataLen = %d\n", datalen);
    DbgPrint(" symInfo.dirPathOffsetLength = %d\n", symInfo.dirPathOffsetLength);
    DbgPrint(" symInfo.linkOffsetLength = %d\n", symInfo.linkOffsetLength);
    DbgPrint(" pdata->datalen = %d\n", pdata->reqLen);

    mydump(sizeof(symInfo), &symInfo);

    cmdlen += datalen;


    cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
    if (cmd)
    {
        cmd->dataLen = datalen;
        cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
        cmd->Command.SequenceNumber = 0;
        cmd->Command.SessionId      = Session;
        cmd->NwcCommand = NWC_MAP_DRIVE;

        copy_from_user(cmd->data, pdata->reqData, datalen);
        status = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);

        if (reply)
        {
            status = reply->Reply.ErrorCode;
            Novfs_Free(reply);
        }
        Novfs_Free(cmd);
    }
    return(status);

}

/*++======================================================================*/
int	NwcUnMapDrive(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract: Get the Primary connection 
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST     cmd;
PXPLAT_CALL_REPLY       reply;
u_long             status = 0, datalen, cmdlen, replylen;
NwcUnmapDriveEx    symInfo;

    DbgPrint("Call to NwcUnMapDrive\n");

    copy_from_user(&symInfo, pdata->reqData, sizeof(symInfo));
    cmdlen = sizeof(*cmd);
    datalen = sizeof(symInfo) + symInfo.linkLen;

    cmdlen += datalen;
    cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
    if (cmd)
    {
        cmd->dataLen = datalen;
        cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
        cmd->Command.SequenceNumber = 0;
        cmd->Command.SessionId      = Session;
        cmd->NwcCommand = NWC_UNMAP_DRIVE;

        copy_from_user(cmd->data, pdata->reqData, datalen);
        status = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);

        if (reply)
        {
            status = reply->Reply.ErrorCode;
            Novfs_Free(reply);
        }
        Novfs_Free(cmd);
    }

    return(status);
}


/*++======================================================================*/
int	NwcEnumerateDrives(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract: Get the Primary connection 
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST     cmd;
PXPLAT_CALL_REPLY       reply;
u_long                  status = 0, cmdlen, replylen;
u_long                  offset;
char                    *cp;

    DbgPrint("Call to NwcEnumerateDrives\n");

    cmdlen = sizeof(*cmd);
    cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
    if (cmd)
    {
        cmd->dataLen = 0;
        cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
        cmd->Command.SequenceNumber = 0;
        cmd->Command.SessionId      = Session;
        cmd->NwcCommand = NWC_ENUMERATE_DRIVES;
        status = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);

        if (reply)
        {
            status = reply->Reply.ErrorCode;
            DbgPrint("Status Code = 0x%X\n", status);
            if (!status)
            {
                offset = sizeof(u_long);
                cp = reply->data;
                replylen = ((PNwcGetMappedDrives)pdata->repData)->MapBuffLen;
                copy_to_user(pdata->repData, cp, offset);
                cp += offset;
                copy_to_user(((PNwcGetMappedDrives)pdata->repData)->MapBuffer, cp, min(replylen - offset, reply->dataLen - offset));
            }

            Novfs_Free(reply);
        }
        Novfs_Free(cmd);
    }

    return(status);
}


/*++======================================================================*/
int	NwcGetBroadcastMessage(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract: Get the Primary connection 
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST             cmd;
PXPLAT_CALL_REPLY               reply;
u_long                          status = 0, cmdlen, replylen;
NwcGetBroadcastNotification     msg;
PNwdCGetBroadcastNotification   dmsg;

    cmdlen = sizeof(*cmd) + sizeof(*dmsg);
    cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);
    if (cmd)
    {

        copy_from_user(&msg, pdata->reqData, sizeof(msg));
        cmd->dataLen = sizeof(*dmsg);
        cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
        cmd->Command.SequenceNumber = 0;
        cmd->Command.SessionId      = Session;

        cmd->NwcCommand = NWC_GET_BROADCAST_MESSAGE;
        dmsg = (PNwdCGetBroadcastNotification)cmd->data;
        dmsg->uConnReference = msg.uConnReference;
        
        status = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);

        if (reply)
        {
            status = reply->Reply.ErrorCode;
            DbgPrint("Status Code = 0x%X\n", status);
            if (!status)
            {
                /* we have a message so copy it to the user buffer */
                copy_to_user(pdata->repData, reply->data, min(pdata->repLen, reply->dataLen));
            }
            else
            {
                msg.messageLen = 0;
                msg.message[0] = 0;
                copy_to_user(pdata->repData, &msg, sizeof(msg));
            }

            Novfs_Free(reply);
        }
        Novfs_Free(cmd);
    }
    return(status);

}


int NwdSetKeyValue(PXPLAT pdata, session_t Session)
{
PXPLAT_CALL_REQUEST     cmd;
PXPLAT_CALL_REPLY	    reply;
NwcSetKey               xplatCall;
PNwdCSetKey             pNewKey;
NwcString               xferStr;
char                    *str;
u_long                  status = -ENOMEM, cmdlen, datalen, replylen;

	copy_from_user(&xplatCall, pdata->reqData, sizeof(NwcSetKey));
    datalen = sizeof(NwdCSetKey) + xplatCall.pObjectName->DataLen + xplatCall.pNewPassword->DataLen;
    cmdlen = sizeof(*cmd) + datalen;
    cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

    if (cmd)
    {
        pNewKey = (PNwdCSetKey)cmd->data;
        cmd->dataLen = datalen;
   		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_SET_KEY;

        pNewKey->ObjectType = xplatCall.ObjectType;
        pNewKey->AuthenticationId = xplatCall.AuthenticationId;
        pNewKey->ConnHandle = xplatCall.ConnHandle;
        str = (char *)pNewKey;

        /*
         * Get the User Name
         */
        str += sizeof(NwdCSetKey);
        copy_from_user(&xferStr, xplatCall.pObjectName, sizeof(NwcString));
        pNewKey->objectNameOffset = sizeof(*pNewKey);
        copy_from_user(str, xferStr.pBuffer, xferStr.DataLen);
        pNewKey->objectNameLen = xferStr.DataLen;

        /*
         * Get the Verify Password
         */
        str += pNewKey->objectNameLen;
        copy_from_user(&xferStr, xplatCall.pNewPassword, sizeof(NwcString));
        pNewKey->newPasswordOffset = pNewKey->objectNameOffset;
        copy_from_user(str, xferStr.pBuffer, xferStr.DataLen);
        pNewKey->newPasswordLen = xferStr.DataLen;

		status = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
        if (reply)
        {
            status = reply->Reply.ErrorCode;
            Novfs_Free(reply);
        }
        memset(cmd, 0, cmdlen);
        Novfs_Free(cmd);
    }

	return(status);
}
 
/*++======================================================================*/
int NwdVerifyKeyValue(PXPLAT pdata, session_t Session)
/*
 *  Arguments:
 *
 *  Returns:
 *
 *  Abstract: Change the password on the server
 *
 *  Notes:
 *
 *  Environment:
 *
 *========================================================================*/
{
PXPLAT_CALL_REQUEST				cmd;
PXPLAT_CALL_REPLY				reply;
NwcVerifyKey			        xplatCall;
PNwdCVerifyKey		            pNewKey;
NwcString						xferStr;
char							*str;
u_long                          status = -ENOMEM, cmdlen, datalen, replylen;

	copy_from_user(&xplatCall, pdata->reqData, sizeof(NwcVerifyKey));

    datalen = sizeof(NwdCVerifyKey) + xplatCall.pDomainName->DataLen + xplatCall.pObjectName->DataLen
        + xplatCall.pVerifyPassword->DataLen;

    cmdlen = sizeof(*cmd) + datalen;
    cmd = Novfs_Malloc(cmdlen, GFP_KERNEL);

    if (cmd)
    {
        pNewKey = (PNwdCVerifyKey)cmd->data;
        cmd->dataLen = datalen;
   		cmd->Command.CommandType    = VFS_COMMAND_XPLAT_CALL;
		cmd->Command.SequenceNumber = 0;
		cmd->Command.SessionId      = Session;
		cmd->NwcCommand = NWC_VERIFY_KEY;

        pNewKey->NameType = xplatCall.NameType;
        pNewKey->ObjectType = xplatCall.ObjectType;
        pNewKey->AuthType = xplatCall.AuthType;
        str = (char *)pNewKey;

        /*
         * Get the tree name
         */
        str += sizeof(*pNewKey);
        copy_from_user(&xferStr, xplatCall.pDomainName, sizeof(NwcString));
        pNewKey->domainNameOffset = sizeof(*pNewKey);
        copy_from_user(str, xferStr.pBuffer, xferStr.DataLen);
        pNewKey->domainNameLen = xferStr.DataLen;

        /*
         * Get the User Name
         */
        str += pNewKey->domainNameLen;
        copy_from_user(&xferStr, xplatCall.pObjectName, sizeof(NwcString));
        pNewKey->objectNameOffset = pNewKey->domainNameOffset + pNewKey->domainNameLen;
        copy_from_user(str, xferStr.pBuffer, xferStr.DataLen);
        pNewKey->objectNameLen = xferStr.DataLen;

        /*
         * Get the Verify Password
         */
        str += pNewKey->objectNameLen;
        copy_from_user(&xferStr, xplatCall.pVerifyPassword, sizeof(NwcString));
        pNewKey->verifyPasswordOffset = pNewKey->objectNameOffset + pNewKey->objectNameLen;
        copy_from_user(str, xferStr.pBuffer, xferStr.DataLen);
        pNewKey->verifyPasswordLen = xferStr.DataLen;

		status = Queue_Daemon_Command((void *)cmd, cmdlen, NULL, 0, (void **)&reply, &replylen, INTERRUPTIBLE);
        if (reply)
        {
            status = reply->Reply.ErrorCode;
            Novfs_Free(reply);
        }
        memset(cmd, 0, cmdlen);
        Novfs_Free(cmd);
    }

	return(status);
}
